'use strict'

Object.defineProperty(exports, '__esModule', { value: true })

var get = require('lodash/get')
var defaultsDeep = require('lodash/defaultsDeep')
var path = require('upath')
var fs = require('fs')
var hooksCore = require('@midwayjs/hooks-core')
var findUp = require('find-up')
var glob = require('@midwayjs/glob')
var kebabCase = require('lodash/kebabCase')
var last = require('lodash/last')
var lodash = require('lodash')
var pickBy = require('lodash/pickBy')
var parseFunctionArgs = require('fn-args')

function _interopDefault(e) {
  return e && e.__esModule ? e : { default: e }
}

var get__default = /*#__PURE__*/ _interopDefault(get)
var defaultsDeep__default = /*#__PURE__*/ _interopDefault(defaultsDeep)
var path__default = /*#__PURE__*/ _interopDefault(path)
var findUp__default = /*#__PURE__*/ _interopDefault(findUp)
var kebabCase__default = /*#__PURE__*/ _interopDefault(kebabCase)
var last__default = /*#__PURE__*/ _interopDefault(last)
var pickBy__default = /*#__PURE__*/ _interopDefault(pickBy)
var parseFunctionArgs__default =
  /*#__PURE__*/ _interopDefault(parseFunctionArgs)

const HOOKS_DEV_MODULE_PATH = 'HOOKS_DEV_MODULE_PATH'
const HOOKS_PROJECT_CONFIG = 'HOOKS_PROJECT_CONFIG'
const HOOKS_PROJECT_ROOT = 'HOOKS_PROJECT_ROOT'

const debug$2 = hooksCore.createDebug('hooks-internal:root')

function setProjectRoot(root) {
  debug$2('setProjectRoot: %s', root)
  process.env[HOOKS_PROJECT_ROOT] = root
}

function getProjectRoot(cwd) {
  if (process.env[HOOKS_PROJECT_ROOT]) {
    return process.env[HOOKS_PROJECT_ROOT]
  }

  const pkg = findUp__default['default'].sync('package.json', { cwd })
  if (pkg) {
    return path.dirname(pkg)
  }

  return process.cwd()
}

const debug$1 = hooksCore.createDebug('hooks-internal: config')

function setConfig(config) {
  debug$1('setConfig: %O', config)
  process.env[HOOKS_PROJECT_CONFIG] = JSON.stringify(config)
}

function getConfig(cwd = getProjectRoot()) {
  const preDefineConfig = process.env[HOOKS_PROJECT_CONFIG]

  const userConfig = preDefineConfig
    ? JSON.parse(preDefineConfig)
    : getConfigFromFile(cwd)

  const deprecatedConfig = ['dev.include', 'dev.exclude', 'dev.ignorePattern']

  for (const key of deprecatedConfig) {
    if (get__default['default'](userConfig, key)) {
      console.warn(
        `[hooks] ${key} is no longer supported, please remove it from your config`
      )
    }
  }

  return defaultsDeep__default['default']({}, userConfig, {
    source: './src/api',
    build: { outDir: './dist' },
  })
}

function getConfigFromFile(root) {
  const configs = [
    path__default['default'].join(root, 'midway.config.ts'),
    path__default['default'].join(root, 'midway.config.js'),
  ]

  const config = configs.find(fs.existsSync)

  if (!config) {
    throw new Error(
      `[ERR_INVALID_CONFIG] midway.config.[ts|js] is not found at ${root}`
    )
  }

  return requireMod(config)
}

function defineConfig(config) {
  return config
}

const requireMod = (id) => {
  const contents = require(id)
  debug$1('requireMod: %s, %O', id, contents)
  if ('default' in contents) return contents.default
  return contents
}

exports.RouteKeyword = void 0
;(function (RouteKeyword) {
  RouteKeyword['INDEX'] = 'index'
  RouteKeyword['CATCH_ALL'] = '/*'
  RouteKeyword['DYNAMIC'] = ':'
})(exports.RouteKeyword || (exports.RouteKeyword = {}))

class FileSystemRouter extends hooksCore.AbstractRouter {
  constructor(config) {
    super()
    this.config = config
  }

  getApiDirectory(baseDir) {
    return path.join(this.config.source, baseDir)
  }

  /**
   * legacy: index-handler
   * now: indexHandler
   */
  getFunctionId(file, functionName, isExportDefault) {
    const relativePath = path.relative(this.config.source, file)
    // src/apis/lambda/index.ts -> apis-lambda-index
    const id = kebabCase__default['default'](
      path.removeExt(relativePath, path.extname(relativePath))
    )
    const name = [id, isExportDefault ? '' : `-${functionName}`].join('')
    return this.config.legacy ? name.toLowerCase() : lodash.camelCase(name)
  }

  getRoute(file) {
    return this.config.routes.find((route) => {
      const apiDir = this.getApiDirectory(route.baseDir)
      return this.isPathInside(path.toUnix(file), path.toUnix(apiDir))
    })
  }

  isApiFile({ file }) {
    const route = this.getRoute(file)
    return !!route
  }

  functionToHttpPath(file, functionName, exportDefault) {
    const { baseDir } = this.getRoute(file)

    const filePath = path.removeExt(
      path.relative(this.getApiDirectory(baseDir), file),
      path.extname(file)
    )

    return hooksCore.urlJoin(
      this.buildUrl(filePath, exportDefault ? '' : functionName),
      {
        trailingSlash: false,
      }
    )
  }

  buildUrl(file, functionName = '') {
    const parts = file
      .split(/\[(.+?\(.+?\)|.+?)\]/)
      .map((str, i) => {
        if (!str) return null
        const dynamic = i % 2 === 1

        const [, content, qualifier] = dynamic
          ? /([^(]+)(\(.+\))?$/.exec(str)
          : [null, str, null]

        return {
          content,
          dynamic,
          catchAll: /^\.{3}.+$/.test(content),
          qualifier,
        }
      })
      .filter(Boolean)

    const catchAllIndex = parts.findIndex((part) => part.catchAll)
    if (catchAllIndex !== -1 && catchAllIndex !== parts.length - 1) {
      throw new Error(
        `Catch all routes must be the last part of the path. Current input: ${file}`
      )
    }

    const segments = []

    for (let [idx, { content, dynamic, catchAll }] of parts.entries()) {
      if (!content) continue

      if (catchAll) {
        if (content.slice(3) !== exports.RouteKeyword.INDEX) {
          segments.push(content.slice(3))
        }
        continue
      }

      /**
       * /[api]/id -> /:api/id
       */
      if (dynamic) {
        segments.push(`${exports.RouteKeyword.DYNAMIC}${content}`)
        continue
      }

      if (idx === parts.length - 1) {
        const contents = content.split('/')
        if (last__default['default'](contents) === exports.RouteKeyword.INDEX) {
          contents.pop()
          segments.push(contents.join('/'))
          continue
        }
      }

      segments.push(content)
    }

    segments.push(functionName)
    if (catchAllIndex !== -1) {
      segments.push(exports.RouteKeyword.CATCH_ALL)
    }

    return hooksCore.urlJoin.apply(null, [...segments, {}])
  }
}

const debug = hooksCore.createDebug('hooks-internal: router')

function getRouter(source) {
  const { routes, legacy } = getConfig()

  if (Array.isArray(routes) && routes.length > 0) {
    return new FileSystemRouter({ source, routes, legacy })
  }

  return new hooksCore.ApiRouter()
}

function getSource(options) {
  const root = getProjectRoot()
  const {
    source,
    build: { outDir },
  } = getConfig()

  return path.join(root, options.useSourceFile ? source : outDir)
}

function normalizePath(router, api) {
  const { trigger, file } = api

  if (router instanceof FileSystemRouter) {
    const basePath = router.getRoute(file).basePath
    return hooksCore.urlJoin(basePath, trigger.path, {})
  }

  return trigger.path
}

function loadApiFiles(options) {
  const { source, router } = options

  debug('load source: %s', source)

  const files = glob.run(['**/*.{ts,js}'], {
    cwd: source,
    ignore: [
      '**/node_modules/**',
      '**/*.d.ts',
      '**/*.{test,spec}.{ts,tsx,js,jsx,mjs}',
      '**/_client/**/*.js',
      '**/configuration.{ts,js}',
    ],
  })

  debug('scan files: %O', files)
  if (files.length === 0) {
    console.warn('No api files found, source is:', source)
  }

  const apis = files.filter(
    (file) =>
      router.isSourceFile(file, source) &&
      router.isApiFile({ file, mod: require(file) })
  )

  debug('api files: %O', apis)

  return {
    apis,
    files,
  }
}

function parseApiModule(mod, file, router) {
  if (router instanceof FileSystemRouter) {
    for (const [name, fn] of Object.entries(
      pickBy__default['default'](mod, hooksCore.isFunction)
    )) {
      if (!Reflect.getMetadata(hooksCore.OperatorType.Trigger, fn)) {
        // default is http
        const HttpMethod =
          parseFunctionArgs__default['default'](fn).length === 0
            ? hooksCore.Get
            : hooksCore.Post
        mod[name] = hooksCore.Api(HttpMethod(), fn)
      }
    }
  }
  return hooksCore.parseApiModule(mod, file, router)
}

function isDev() {
  return (
    process.env.MIDWAY_TS_MODE === 'true' ||
    /* istanbul ignore next */
    process.env.NODE_ENV === 'test' ||
    /* istanbul ignore next */
    process.env.NODE_ENV === 'development'
  )
}

const MIDWAY_HOOKS_HYDRATE_OPTIONS = 'MIDWAY_HOOKS_HYDRATE_OPTIONS'

function isHydrate() {
  return !!globalThis[MIDWAY_HOOKS_HYDRATE_OPTIONS]
}

function setHydrateOptions(options) {
  setProjectRoot('/')
  setConfig({})
  globalThis[MIDWAY_HOOKS_HYDRATE_OPTIONS] = options
}

function getHydrateOptions() {
  return globalThis[MIDWAY_HOOKS_HYDRATE_OPTIONS]
}

exports.FileSystemRouter = FileSystemRouter
exports.HOOKS_DEV_MODULE_PATH = HOOKS_DEV_MODULE_PATH
exports.HOOKS_PROJECT_CONFIG = HOOKS_PROJECT_CONFIG
exports.HOOKS_PROJECT_ROOT = HOOKS_PROJECT_ROOT
exports.defineConfig = defineConfig
exports.getConfig = getConfig
exports.getConfigFromFile = getConfigFromFile
exports.getHydrateOptions = getHydrateOptions
exports.getProjectRoot = getProjectRoot
exports.getRouter = getRouter
exports.getSource = getSource
exports.isDev = isDev
exports.isHydrate = isHydrate
exports.loadApiFiles = loadApiFiles
exports.normalizePath = normalizePath
exports.parseApiModule = parseApiModule
exports.setConfig = setConfig
exports.setHydrateOptions = setHydrateOptions
exports.setProjectRoot = setProjectRoot
