import get from 'lodash/get'
import defaultsDeep from 'lodash/defaultsDeep'
import path, {
  dirname,
  join,
  relative,
  removeExt,
  extname,
  toUnix,
} from 'upath'
import { existsSync } from 'fs'
import {
  createDebug,
  AbstractRouter,
  urlJoin,
  ApiRouter,
  isFunction,
  OperatorType,
  Get,
  Post,
  Api,
  parseApiModule as parseApiModule$1,
} from '@midwayjs/hooks-core'
import findUp from 'find-up'
import { run } from '@midwayjs/glob'
import kebabCase from 'lodash/kebabCase'
import last from 'lodash/last'
import { camelCase } from 'lodash'
import pickBy from 'lodash/pickBy'
import parseFunctionArgs from 'fn-args'

const HOOKS_DEV_MODULE_PATH = 'HOOKS_DEV_MODULE_PATH'
const HOOKS_PROJECT_CONFIG = 'HOOKS_PROJECT_CONFIG'
const HOOKS_PROJECT_ROOT = 'HOOKS_PROJECT_ROOT'

const debug$2 = createDebug('hooks-internal:root')

function setProjectRoot(root) {
  debug$2('setProjectRoot: %s', root)
  process.env[HOOKS_PROJECT_ROOT] = root
}

function getProjectRoot(cwd) {
  if (process.env[HOOKS_PROJECT_ROOT]) {
    return process.env[HOOKS_PROJECT_ROOT]
  }

  const pkg = findUp.sync('package.json', { cwd })
  if (pkg) {
    return dirname(pkg)
  }

  return process.cwd()
}

const debug$1 = createDebug('hooks-internal: config')

function setConfig(config) {
  debug$1('setConfig: %O', config)
  process.env[HOOKS_PROJECT_CONFIG] = JSON.stringify(config)
}

function getConfig(cwd = getProjectRoot()) {
  const preDefineConfig = process.env[HOOKS_PROJECT_CONFIG]

  const userConfig = preDefineConfig
    ? JSON.parse(preDefineConfig)
    : getConfigFromFile(cwd)

  const deprecatedConfig = ['dev.include', 'dev.exclude', 'dev.ignorePattern']

  for (const key of deprecatedConfig) {
    if (get(userConfig, key)) {
      console.warn(
        `[hooks] ${key} is no longer supported, please remove it from your config`
      )
    }
  }

  return defaultsDeep({}, userConfig, {
    source: './src/api',
    build: { outDir: './dist' },
  })
}

function getConfigFromFile(root) {
  const configs = [
    path.join(root, 'midway.config.ts'),
    path.join(root, 'midway.config.js'),
  ]

  const config = configs.find(existsSync)

  if (!config) {
    throw new Error(
      `[ERR_INVALID_CONFIG] midway.config.[ts|js] is not found at ${root}`
    )
  }

  return requireMod(config)
}

function defineConfig(config) {
  return config
}

const requireMod = (id) => {
  const contents = require(id)
  debug$1('requireMod: %s, %O', id, contents)
  if ('default' in contents) return contents.default
  return contents
}

let RouteKeyword
;(function (RouteKeyword) {
  RouteKeyword['INDEX'] = 'index'
  RouteKeyword['CATCH_ALL'] = '/*'
  RouteKeyword['DYNAMIC'] = ':'
})(RouteKeyword || (RouteKeyword = {}))

class FileSystemRouter extends AbstractRouter {
  constructor(config) {
    super()
    this.config = config
  }

  getApiDirectory(baseDir) {
    return join(this.config.source, baseDir)
  }

  /**
   * legacy: index-handler
   * now: indexHandler
   */
  getFunctionId(file, functionName, isExportDefault) {
    const relativePath = relative(this.config.source, file)
    // src/apis/lambda/index.ts -> apis-lambda-index
    const id = kebabCase(removeExt(relativePath, extname(relativePath)))
    const name = [id, isExportDefault ? '' : `-${functionName}`].join('')
    return this.config.legacy ? name.toLowerCase() : camelCase(name)
  }

  getRoute(file) {
    return this.config.routes.find((route) => {
      const apiDir = this.getApiDirectory(route.baseDir)
      return this.isPathInside(toUnix(file), toUnix(apiDir))
    })
  }

  isApiFile({ file }) {
    const route = this.getRoute(file)
    return !!route
  }

  functionToHttpPath(file, functionName, exportDefault) {
    const { baseDir } = this.getRoute(file)

    const filePath = removeExt(
      relative(this.getApiDirectory(baseDir), file),
      extname(file)
    )

    return urlJoin(this.buildUrl(filePath, exportDefault ? '' : functionName), {
      trailingSlash: false,
    })
  }

  buildUrl(file, functionName = '') {
    const parts = file
      .split(/\[(.+?\(.+?\)|.+?)\]/)
      .map((str, i) => {
        if (!str) return null
        const dynamic = i % 2 === 1

        const [, content, qualifier] = dynamic
          ? /([^(]+)(\(.+\))?$/.exec(str)
          : [null, str, null]

        return {
          content,
          dynamic,
          catchAll: /^\.{3}.+$/.test(content),
          qualifier,
        }
      })
      .filter(Boolean)

    const catchAllIndex = parts.findIndex((part) => part.catchAll)
    if (catchAllIndex !== -1 && catchAllIndex !== parts.length - 1) {
      throw new Error(
        `Catch all routes must be the last part of the path. Current input: ${file}`
      )
    }

    const segments = []

    for (let [idx, { content, dynamic, catchAll }] of parts.entries()) {
      if (!content) continue

      if (catchAll) {
        if (content.slice(3) !== RouteKeyword.INDEX) {
          segments.push(content.slice(3))
        }
        continue
      }

      /**
       * /[api]/id -> /:api/id
       */
      if (dynamic) {
        segments.push(`${RouteKeyword.DYNAMIC}${content}`)
        continue
      }

      if (idx === parts.length - 1) {
        const contents = content.split('/')
        if (last(contents) === RouteKeyword.INDEX) {
          contents.pop()
          segments.push(contents.join('/'))
          continue
        }
      }

      segments.push(content)
    }

    segments.push(functionName)
    if (catchAllIndex !== -1) {
      segments.push(RouteKeyword.CATCH_ALL)
    }

    return urlJoin.apply(null, [...segments, {}])
  }
}

const debug = createDebug('hooks-internal: router')

function getRouter(source) {
  const { routes, legacy } = getConfig()

  if (Array.isArray(routes) && routes.length > 0) {
    return new FileSystemRouter({ source, routes, legacy })
  }

  return new ApiRouter()
}

function getSource(options) {
  const root = getProjectRoot()
  const {
    source,
    build: { outDir },
  } = getConfig()

  return join(root, options.useSourceFile ? source : outDir)
}

function normalizePath(router, api) {
  const { trigger, file } = api

  if (router instanceof FileSystemRouter) {
    const basePath = router.getRoute(file).basePath
    return urlJoin(basePath, trigger.path, {})
  }

  return trigger.path
}

function loadApiFiles(options) {
  const { source, router } = options

  debug('load source: %s', source)

  const files = run(['**/*.{ts,js}'], {
    cwd: source,
    ignore: [
      '**/node_modules/**',
      '**/*.d.ts',
      '**/*.{test,spec}.{ts,tsx,js,jsx,mjs}',
      '**/_client/**/*.js',
      '**/configuration.{ts,js}',
    ],
  })

  debug('scan files: %O', files)
  if (files.length === 0) {
    console.warn('No api files found, source is:', source)
  }

  const apis = files.filter(
    (file) =>
      router.isSourceFile(file, source) &&
      router.isApiFile({ file, mod: require(file) })
  )

  debug('api files: %O', apis)

  return {
    apis,
    files,
  }
}

function parseApiModule(mod, file, router) {
  if (router instanceof FileSystemRouter) {
    for (const [name, fn] of Object.entries(pickBy(mod, isFunction))) {
      if (!Reflect.getMetadata(OperatorType.Trigger, fn)) {
        // default is http
        const HttpMethod = parseFunctionArgs(fn).length === 0 ? Get : Post
        mod[name] = Api(HttpMethod(), fn)
      }
    }
  }
  return parseApiModule$1(mod, file, router)
}

function isDev() {
  return (
    process.env.MIDWAY_TS_MODE === 'true' ||
    /* istanbul ignore next */
    process.env.NODE_ENV === 'test' ||
    /* istanbul ignore next */
    process.env.NODE_ENV === 'development'
  )
}

const MIDWAY_HOOKS_HYDRATE_OPTIONS = 'MIDWAY_HOOKS_HYDRATE_OPTIONS'

function isHydrate() {
  return !!globalThis[MIDWAY_HOOKS_HYDRATE_OPTIONS]
}

function setHydrateOptions(options) {
  setProjectRoot('/')
  setConfig({})
  globalThis[MIDWAY_HOOKS_HYDRATE_OPTIONS] = options
}

function getHydrateOptions() {
  return globalThis[MIDWAY_HOOKS_HYDRATE_OPTIONS]
}

export {
  FileSystemRouter,
  HOOKS_DEV_MODULE_PATH,
  HOOKS_PROJECT_CONFIG,
  HOOKS_PROJECT_ROOT,
  RouteKeyword,
  defineConfig,
  getConfig,
  getConfigFromFile,
  getHydrateOptions,
  getProjectRoot,
  getRouter,
  getSource,
  isDev,
  isHydrate,
  loadApiFiles,
  normalizePath,
  parseApiModule,
  setConfig,
  setHydrateOptions,
  setProjectRoot,
}
